import json
import os
from collections import defaultdict

import requests
import yaml
from fastapi import Depends, FastAPI, HTTPException, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse, HTMLResponse
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from frozendict import frozendict
from qubed import Qube
from qubed.tree_formatters import node_tree_to_html

app = FastAPI()
security = HTTPBearer()
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.mount("/static", StaticFiles(directory="static"), name="static")
templates = Jinja2Templates(directory="templates")

qubes: dict[str, Qube] = {}
# print("Getting climate and extremes dt data from github")
# try:
#     qubes["climate-dt"] = Qube.from_json(
#         requests.get(
#             "https://github.com/ecmwf/qubed/raw/refs/heads/main/tests/example_qubes/climate_dt.json",
#         timeout=3).json()
#     )
#     qubes["extremes-dt"] = Qube.from_json(
#         requests.get(
#             "https://github.com/ecmwf/qubed/raw/refs/heads/main/tests/example_qubes/extremes_dt.json",
#         timeout=3).json()
#     )
#     mars_language = yaml.safe_load(
#         requests.get(
#             "https://github.com/ecmwf/qubed/raw/refs/heads/main/config/climate-dt/language.yaml",
#         timeout=3).content
#     )
# except:
qubes["climate-dt"] = Qube.empty()
qubes["extremes-dt"] = Qube.empty()
mars_language = {}

if "LOCAL_CACHE" in os.environ:
    print("Getting climate and extremes dt data from local files")
    with open("../tests/example_qubes/climate_dt.json") as f:
        qubes["climate-dt"] = Qube.from_json(json.load(f))

    with open("../tests/example_qubes/extremes_dt.json") as f:
        qubes["climate-dt"] = qubes["climate-dt"] | Qube.from_json(json.load(f))

    with open("../config/climate-dt/language.yaml", "r") as f:
        mars_language = yaml.safe_load(f)["_field"]
else:
    print("Getting climate and extremes dt data from github")
    qubes["climate-dt"] = Qube.from_json(
        requests.get(
            "https://github.com/ecmwf/qubed/raw/refs/heads/main/tests/example_qubes/climate_dt.json",
            timeout=1,
        ).json()
    )
    qubes["extremes-dt"] = Qube.from_json(
        requests.get(
            "https://github.com/ecmwf/qubed/raw/refs/heads/main/tests/example_qubes/extremes_dt.json",
            timeout=1,
        ).json()
    )

    qubes["od"] = Qube.from_json(
        requests.get(
            "https://github.com/ecmwf/qubed/raw/refs/heads/main/tests/example_qubes/od.json",
            timeout=1,
        ).json()
    )
    qubes["climate-dt"] = qubes["climate-dt"] | qubes["extremes-dt"] | qubes["od"]
    mars_language = yaml.safe_load(
        requests.get(
            "https://github.com/ecmwf/qubed/raw/refs/heads/main/config/climate-dt/language.yaml",
            timeout=3,
        ).content
    )["_field"]

if "API_KEY" in os.environ:
    api_key = os.environ["API_KEY"]
else:
    with open("api_key.secret", "r") as f:
        api_key = f.read()

print("Ready to serve requests!")


def validate_key(key: str):
    if key not in qubes:
        raise HTTPException(status_code=404, detail=f"Qube {key} not found")
    return key


async def get_body_json(request: Request):
    return await request.json()


def parse_request(request: Request) -> dict[str, str | list[str]]:
    # Convert query parameters to dictionary format
    request_dict = dict(request.query_params)
    for key, value in request_dict.items():
        # Convert comma-separated values into lists
        if "," in value:
            request_dict[key] = value.split(",")

    return request_dict


def validate_api_key(credentials: HTTPAuthorizationCredentials = Depends(security)):
    if credentials.credentials != api_key:
        raise HTTPException(status_code=403, detail="Incorrect API Key")
    return credentials


@app.get("/favicon.ico", include_in_schema=False)
async def favicon():
    return FileResponse("favicon.ico")


@app.get("/", response_class=HTMLResponse)
async def read_root(request: Request):
    return templates.TemplateResponse(
        "index.html",
        {
            "request": request,
            "config": {
                "message": "Hello from the dev server!",
            },
            "api_url": os.environ.get("API_URL", "/api/v1/"),
        },
    )


@app.get("/api/v1/keys/")
async def keys():
    return list(qubes.keys())


@app.get("/api/v1/get/{key}/")
async def get(
    key: str = Depends(validate_key),
    request: dict[str, str | list[str]] = Depends(parse_request),
):
    return qubes[key].to_json()


@app.post("/api/v1/union/{key}/")
async def union(
    key: str,
    credentials: HTTPAuthorizationCredentials = Depends(validate_api_key),
    body_json=Depends(get_body_json),
):
    if key not in qubes:
        qubes[key] = Qube.empty()

    q = Qube.from_json(body_json)
    qubes[key] = qubes[key] | q
    return qubes[key].to_json()


def follow_query(request: dict[str, str | list[str]], qube: Qube):
    s = qube.select(request, mode="next_level", prune=True, consume=False)
    by_path = defaultdict(lambda: {"paths": set(), "values": set()})

    for request, node in s.leaf_nodes():
        if not node.data.metadata["is_leaf"]:
            by_path[node.key]["values"].update(node.values.values)
            by_path[node.key]["paths"].add(frozendict(request))

    return s, [
        {
            "paths": list(v["paths"]),
            "key": key,
            "values": sorted(v["values"], reverse=True),
        }
        for key, v in by_path.items()
    ]


@app.get("/api/v1/select/{key}/")
async def select(
    key: str = Depends(validate_key),
    request: dict[str, str | list[str]] = Depends(parse_request),
):
    q = qubes[key].select(request)
    return q.to_json()


@app.get("/api/v1/query/{key}")
async def query(
    key: str = Depends(validate_key),
    request: dict[str, str | list[str]] = Depends(parse_request),
):
    qube, paths = follow_query(request, qubes[key])
    return paths


@app.get("/api/v1/basicstac/{key}/{filters:path}")
async def basic_stac(filters: str, key: str = Depends(validate_key)):
    pairs = filters.strip("/").split("/")
    request = dict(p.split("=") for p in pairs if "=" in p)

    qube, _ = follow_query(request, qubes[key])

    def make_link(child_request):
        """Take a MARS Key and information about which paths matched up to this point and use it to make a STAC Link"""
        kvs = [f"{key}={value}" for key, value in child_request.items()]
        href = f"/api/v1/basicstac/{key}/{'/'.join(kvs)}"
        last_key, last_value = list(child_request.items())[-1]

        return {
            "title": f"{last_key}={last_value}",
            "href": href,
            "rel": "child",
            "type": "application/json",
        }

    # Format the response as a STAC collection
    (this_key, this_value), *_ = (
        list(request.items())[-1] if request else ("root", "root"),
        None,
    )
    key_info = mars_language.get(this_key, {})
    try:
        values_info = dict(key_info.get("values", {}))
        value_info = values_info.get(
            this_value, f"No info found for value `{this_value}` found."
        )
    except ValueError:
        value_info = f"No info found for value `{this_value}` found."

    if this_key == "root":
        value_info = "The root node"
    # key_desc = key_info.get(
    #     "description", f"No description for `key` {this_key} found."
    # )
    print(this_key, this_value)

    print(this_key, key_info)
    stac_collection = {
        "type": "Catalog",
        "stac_version": "1.0.0",
        "id": "root"
        if not request
        else "/".join(f"{k}={v}" for k, v in request.items()),
        "title": f"{this_key}={this_value}",
        "description": value_info,
        "links": [make_link(leaf) for leaf in qube.leaves()],
        # "debug": {
        #     "qube": str(qube),
        # },
    }

    return stac_collection


@app.get("/api/v1/stac/{key}/")
async def get_STAC(
    key: str = Depends(validate_key),
    request: dict[str, str | list[str]] = Depends(parse_request),
):
    qube, paths = follow_query(request, qubes[key])
    kvs = [
        f"{k}={','.join(v)}" if isinstance(v, list) else f"{k}={v}"
        for k, v in request.items()
    ]
    request_params = "&".join(kvs)

    def make_link(key_name, paths, values):
        """Take a MARS Key and information about which paths matched up to this point and use it to make a STAC Link"""
        href_template = f"/stac?{request_params}{'&' if request_params else ''}{key_name}={{{key_name}}}"
        values_from_mars_language = mars_language.get(key_name, {}).get("values", [])

        if all(isinstance(v, list) for v in values_from_mars_language):
            value_descriptions_dict = {
                k: v[-1]
                for v in values_from_mars_language
                if len(v) > 1
                for k in v[:-1]
            }
            value_descriptions = [value_descriptions_dict.get(v, "") for v in values]
            if not any(value_descriptions):
                value_descriptions = None

        return {
            "title": key_name,
            "uriTemplate": href_template,
            "rel": "child",
            "type": "application/json",
            "variables": {
                key_name: {
                    "type": "string",
                    "description": mars_language.get(key_name, {}).get(
                        "description", ""
                    ),
                    "enum": values,
                    "value_descriptions": value_descriptions,
                    # "paths": paths,
                }
            },
        }

    def value_descriptions(key, values):
        return {
            v[0]: v[-1]
            for v in mars_language.get(key, {}).get("values", [])
            if len(v) > 1 and v[0] in list(values)
        }

    descriptions = {
        key: {
            "key": key,
            "values": values,
            "description": mars_language.get(key, {}).get("description", ""),
            "value_descriptions": value_descriptions(key, values),
        }
        for key, values in request.items()
    }

    # Format the response as a STAC collection
    stac_collection = {
        "type": "Catalog",
        "stac_version": "1.0.0",
        "id": "root" if not request else "/stac?" + request_params,
        "description": "STAC collection representing potential children of this request",
        "links": [make_link(p["key"], p["paths"], p["values"]) for p in paths],
        "debug": {
            # "request": request,
            "descriptions": descriptions,
            # "paths": paths,
            "qube": node_tree_to_html(
                qube.compress(),
                collapse=True,
                depth=10,
                include_css=False,
                include_js=False,
                max_summary_length=200,
                css_id="qube",
            ),
        },
    }

    return stac_collection



from qubed import Qube

q = Qube.from_dict(
    {
        "class=od": {
            "expver=0001": {"param=1": {}, "param=2": {}},
            "expver=0002": {"param=1": {}, "param=2": {}},
        },
        "class=rd": {
            "expver=0001": {"param=1": {}, "param=2": {}, "param=3": {}},
            "expver=0002": {"param=1": {}, "param=2": {}},
        },
    }
)

wild_datacube = {
    "class": "*",
    "expver": "*",
    "param": "1",
}


def test_wildcard_creation():
    Qube.from_datacube(wild_datacube)


def test_intersection():
    wild_qube = Qube.from_datacube(wild_datacube)
    intersection = q & wild_qube
    assert intersection == Qube.from_dict(
        {
            "class=od/rd": {
                "expver=0001/0002": {"param=1": {}},
            },
        }
    )



from qubed import Qube

q = Qube.from_dict(
    {
        "class=od": {
            "expver=0001": {"param=1": {}, "param=2": {}},
            "expver=0002": {"param=1": {}, "param=2": {}},
        },
        "class=rd": {"param=1": {}, "param=2": {}, "param=3": {}},
    }
)


def test_consumption():
    assert q.select({"expver": "0001"}, consume=True) == Qube.from_dict(
        {"class=od": {"expver=0001": {"param=1": {}, "param=2": {}}}}
    )


def test_consumption_off():
    expected = Qube.from_dict(
        {
            "class=od": {"expver=0001": {"param=1": {}, "param=2": {}}},
            "class=rd": {"param=1": {}, "param=2": {}, "param=3": {}},
        }
    )
    assert q.select({"expver": "0001"}, consume=False) == expected


def test_function_input_to_select():
    q = Qube.from_tree("""
    root, frequency=6:00:00
    ├── levtype=pl, param=t, levelist=850, threshold=-2/-4/-8/2/4/8
    └── levtype=sfc
        ├── param=10u/10v, threshold=10/15
        ├── param=2t, threshold=273.15
        └── param=tp, threshold=0.1/1/10/100/20/25/5/50
    """).convert_dtypes(
        {
            "threshold": float,
        }
    )

    r = q.select(
        {
            "threshold": lambda t: t > 5,
        }
    )

    assert r == Qube.from_tree("""
    root, frequency=6:00:00
    ├── levtype=pl, param=t, levelist=850, threshold=8
    └── levtype=sfc
        ├── param=10u/10v, threshold=10/15
        ├── param=2t, threshold=273.15
        └── param=tp, threshold=10/100/20/25/50
    """).convert_dtypes(
        {
            "threshold": float,
        }
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "qubed"
copyright = "2025, Tom Hodson (ECMWF)"
author = "Tom Hodson (ECMWF)"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",  # for generating documentation from the docstrings in our code
    "sphinx.ext.napoleon",  # for parsing Numpy and Google stye docstrings
    "myst_nb",  # For parsing markdown
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "jupyter_execute"]


source_suffix = {
    ".rst": "restructuredtext",
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

myst_enable_extensions = [
    "attrs_inline",
]



from frozendict import frozendict
from qubed import Qube


def test_iter_leaves_simple():
    def make_hashable(list_like):
        for d in list_like:
            yield frozendict(d)

    q = Qube.from_dict({"a=1/2": {"b=1/2": {}}})
    entries = [
        {"a": "1", "b": "1"},
        {"a": "1", "b": "2"},
        {"a": "2", "b": "1"},
        {"a": "2", "b": "2"},
    ]

    assert set(make_hashable(q.leaves())) == set(make_hashable(entries))



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass, replace
from datetime import date, datetime, timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    FrozenSet,
    Iterable,
    Iterator,
    Literal,
    Sequence,
    TypeVar,
)

if TYPE_CHECKING:
    from .Qube import Qube


@dataclass(frozen=True)
class ValueGroup(ABC):
    @abstractmethod
    def summary(self) -> str:
        "Provide a string summary of the value group."
        pass

    @abstractmethod
    def __contains__(self, value: Any) -> bool:
        "Given a value, coerce to the value type and determine if it is in the value group."
        pass

    @abstractmethod
    def to_json(self) -> dict:
        "Return a JSON serializable representation of the value group."
        pass

    @abstractmethod
    def min(self):
        "Return the minimum value in the group."
        pass

    @classmethod
    @abstractmethod
    def from_strings(cls, values: Iterable[str]) -> Sequence[ValueGroup]:
        "Given a list of strings, return a one or more ValueGroups of this type."
        pass

    @abstractmethod
    def __iter__(self) -> Iterator:
        "Iterate over the values in the group."
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass


T = TypeVar("T")
EnumValuesType = FrozenSet[T]


@dataclass(frozen=True, order=True)
class QEnum(ValueGroup):
    """
    The simplest kind of key value is just a list of strings.
    summary -> string1/string2/string....
    """

    values: EnumValuesType

    def __init__(self, obj):
        object.__setattr__(self, "values", frozenset(obj))

    def __post_init__(self):
        assert isinstance(self.values, tuple)

    def __iter__(self):
        return iter(self.values)

    def __len__(self) -> int:
        return len(self.values)

    def summary(self) -> str:
        return "/".join(map(str, sorted(self.values)))

    def __contains__(self, value: Any) -> bool:
        return value in self.values

    @classmethod
    def from_strings(cls, values: Iterable[str]) -> Sequence[ValueGroup]:
        return [cls(tuple(values))]

    def min(self):
        return min(self.values)

    def to_json(self):
        return list(self.values)


@dataclass(frozen=True, order=True)
class WildcardGroup(ValueGroup):
    def summary(self) -> str:
        return "*"

    def __contains__(self, value: Any) -> bool:
        return True

    def to_json(self):
        return "*"

    def min(self):
        return "*"

    def __len__(self):
        return None

    def __iter__(self):
        return ["*"]

    @classmethod
    def from_strings(cls, values: Iterable[str]) -> Sequence[ValueGroup]:
        return [WildcardGroup()]


class DateEnum(QEnum):
    def summary(self) -> str:
        def fmt(d):
            return d.strftime("%Y%m%d")

        return "/".join(map(fmt, sorted(self.values)))


@dataclass(frozen=True)
class Range(ValueGroup, ABC):
    dtype: str = dataclasses.field(kw_only=True)

    start: Any
    end: Any
    step: Any

    def min(self):
        return self.start

    def __iter__(self) -> Iterator[Any]:
        i = self.start
        while i <= self.end:
            yield i
            i += self.step

    def to_json(self):
        return dataclasses.asdict(self)


@dataclass(frozen=True)
class DateRange(Range):
    start: date
    end: date
    step: timedelta
    dtype: Literal["date"] = dataclasses.field(kw_only=True, default="date")

    def __len__(self) -> int:
        return (self.end - self.start) // self.step

    def __iter__(self) -> Iterator[date]:
        current = self.start
        while current <= self.end if self.step.days > 0 else current >= self.end:
            yield current
            current += self.step

    @classmethod
    def from_strings(cls, values: Iterable[str]) -> Sequence[DateRange | DateEnum]:
        dates = sorted([datetime.strptime(v, "%Y%m%d") for v in values])
        if len(dates) < 2:
            return [DateEnum(dates)]

        ranges: list[DateEnum | DateRange] = []
        current_group, dates = (
            [
                dates[0],
            ],
            dates[1:],
        )
        current_type: Literal["enum", "range"] = "enum"
        while len(dates) > 1:
            if current_type == "range":
                # If the next date fits then add it to the current range
                if dates[0] - current_group[-1] == timedelta(days=1):
                    current_group.append(dates.pop(0))

                # Emit the current range and start a new one
                else:
                    if len(current_group) == 1:
                        ranges.append(DateEnum(current_group))
                    else:
                        ranges.append(
                            DateRange(
                                start=current_group[0],
                                end=current_group[-1],
                                step=timedelta(days=1),
                            )
                        )
                    current_group = [
                        dates.pop(0),
                    ]
                    current_type = "enum"

            if current_type == "enum":
                # If the next date is one more than the last then switch to range mode
                if dates[0] - current_group[-1] == timedelta(days=1):
                    last = current_group.pop()
                    if current_group:
                        ranges.append(DateEnum(current_group))
                    current_group = [last, dates.pop(0)]
                    current_type = "range"

                else:
                    current_group.append(dates.pop(0))

        # Handle remaining `current_group`
        if current_group:
            if current_type == "range":
                ranges.append(
                    DateRange(
                        start=current_group[0],
                        end=current_group[-1],
                        step=timedelta(days=1),
                    )
                )
            else:
                ranges.append(DateEnum(current_group))

        return ranges

    def __contains__(self, value: Any) -> bool:
        v = datetime.strptime(value, "%Y%m%d").date()
        return self.start <= v <= self.end and (v - self.start) % self.step == 0

    def summary(self) -> str:
        def fmt(d):
            return d.strftime("%Y%m%d")

        if self.step == timedelta(days=0):
            return f"{fmt(self.start)}"
        if self.step == timedelta(days=1):
            return f"{fmt(self.start)}/to/{fmt(self.end)}"

        return (
            f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step // timedelta(days=1)}"
        )


@dataclass(frozen=True)
class TimeRange(Range):
    start: int
    end: int
    step: int
    dtype: Literal["time"] = dataclasses.field(kw_only=True, default="time")

    def min(self):
        return self.start

    def __iter__(self) -> Iterator[Any]:
        return super().__iter__()

    @classmethod
    def from_strings(self, values: Iterable[str]) -> list["TimeRange"]:
        times = sorted([int(v) for v in values])
        if len(times) < 2:
            return [TimeRange(start=times[0], end=times[0], step=100)]

        ranges = []
        current_range, times = (
            [
                times[0],
            ],
            times[1:],
        )
        while len(times) > 1:
            if times[0] - current_range[-1] == 1:
                current_range.append(times.pop(0))

            elif len(current_range) == 1:
                ranges.append(
                    TimeRange(start=current_range[0], end=current_range[0], step=0)
                )
                current_range = [
                    times.pop(0),
                ]

            else:
                ranges.append(
                    TimeRange(start=current_range[0], end=current_range[-1], step=1)
                )
                current_range = [
                    times.pop(0),
                ]
        return ranges

    def __len__(self) -> int:
        return (self.end - self.start) // self.step

    def summary(self) -> str:
        def fmt(d):
            return f"{d:04d}"

        if self.step == 0:
            return f"{fmt(self.start)}"
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"

    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0


@dataclass(frozen=True)
class IntRange(Range):
    start: int
    end: int
    step: int
    dtype: Literal["int"] = dataclasses.field(kw_only=True, default="int")

    def __len__(self) -> int:
        return (self.end - self.start) // self.step

    def summary(self) -> str:
        def fmt(d):
            return d

        if self.step == 0:
            return f"{fmt(self.start)}"
        return f"{fmt(self.start)}/to/{fmt(self.end)}/by/{self.step}"

    def __contains__(self, value: Any) -> bool:
        v = int(value)
        return self.start <= v <= self.end and (v - self.start) % self.step == 0

    @classmethod
    def from_strings(self, values: Iterable[str]) -> list["IntRange"]:
        ints = sorted([int(v) for v in values])
        if len(ints) < 2:
            return [IntRange(start=ints[0], end=ints[0], step=0)]

        ranges = []
        current_range, ints = (
            [
                ints[0],
            ],
            ints[1:],
        )
        while len(ints) > 1:
            if ints[0] - current_range[-1] == 1:
                current_range.append(ints.pop(0))

            elif len(current_range) == 1:
                ranges.append(
                    IntRange(start=current_range[0], end=current_range[0], step=0)
                )
                current_range = [
                    ints.pop(0),
                ]

            else:
                ranges.append(
                    IntRange(start=current_range[0], end=current_range[-1], step=1)
                )
                current_range = [
                    ints.pop(0),
                ]
        return ranges


def values_from_json(obj) -> ValueGroup:
    if isinstance(obj, list):
        return QEnum(tuple(obj))

    match obj["dtype"]:
        case "date":
            return DateRange(**obj)
        case "time":
            return TimeRange(**obj)
        case "int":
            return IntRange(**obj)
        case _:
            raise ValueError(f"Unknown dtype {obj['dtype']}")


def convert_datatypes(q: "Qube", conversions: dict[str, ValueGroup]) -> "Qube":
    def _convert(q: "Qube") -> Iterator["Qube"]:
        if q.key in conversions:
            data_type = conversions[q.key]
            assert isinstance(q.values, QEnum), (
                "Only QEnum values can be converted to other datatypes."
            )
            for values_group in data_type.from_strings(q.values):
                # print(values_group)
                yield replace(q, data=replace(q.data, values=values_group))
        else:
            yield q

    return q.transform(_convert)



# This causes python types to be evaluated later,
# allowing you to reference types like Qube inside the definion of the Qube class
# without having to do "Qube"
from __future__ import annotations

import dataclasses
import functools
import json
from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any, Iterable, Iterator, Literal, Mapping, Sequence

import numpy as np
from frozendict import frozendict

from . import set_operations
from .node_types import NodeData, RootNodeData
from .tree_formatters import (
    HTML,
    node_tree_to_html,
    node_tree_to_string,
)
from .value_types import (
    QEnum,
    ValueGroup,
    WildcardGroup,
    values_from_json,
)


@dataclass(frozen=False, eq=True, order=True, unsafe_hash=True)
class Qube:
    data: NodeData
    children: tuple[Qube, ...]

    @property
    def key(self) -> str:
        return self.data.key

    @property
    def values(self) -> ValueGroup:
        return self.data.values

    @property
    def metadata(self) -> Mapping[str, np.ndarray]:
        return self.data.metadata

    def replace(self, **kwargs) -> Qube:
        data_keys = {
            k: v for k, v in kwargs.items() if k in ["key", "values", "metadata"]
        }
        node_keys = {k: v for k, v in kwargs.items() if k == "children"}
        if not data_keys and not node_keys:
            return self
        if not data_keys:
            return dataclasses.replace(self, **node_keys)

        return dataclasses.replace(
            self, data=dataclasses.replace(self.data, **data_keys), **node_keys
        )

    def summary(self) -> str:
        return self.data.summary()

    @classmethod
    def make(cls, key: str, values: ValueGroup, children, **kwargs) -> Qube:
        return cls(
            data=NodeData(key, values, metadata=kwargs.get("metadata", frozendict())),
            children=tuple(sorted(children, key=lambda n: ((n.key, n.values.min())))),
        )

    @classmethod
    def root_node(cls, children: Iterable[Qube]) -> Qube:
        return cls.make("root", QEnum(("root",)), children)

    @classmethod
    def load(cls, path: str | Path) -> Qube:
        with open(path, "r") as f:
            return Qube.from_json(json.load(f))

    @classmethod
    def from_datacube(cls, datacube: dict[str, str | Sequence[str]]) -> Qube:
        key_vals = list(datacube.items())[::-1]

        children: list[Qube] = []
        for key, values in key_vals:
            values_group: ValueGroup
            if values == "*":
                values_group = WildcardGroup()
            elif isinstance(values, list):
                values_group = QEnum(values)
            else:
                values_group = QEnum([values])

            children = [cls.make(key, values_group, children)]

        return cls.root_node(children)

    @classmethod
    def from_json(cls, json: dict) -> Qube:
        def from_json(json: dict) -> Qube:
            return Qube.make(
                key=json["key"],
                values=values_from_json(json["values"]),
                metadata=frozendict(json["metadata"]) if "metadata" in json else {},
                children=(from_json(c) for c in json["children"]),
            )

        return from_json(json)

    def to_json(self) -> dict:
        def to_json(node: Qube) -> dict:
            return {
                "key": node.key,
                "values": node.values.to_json(),
                "metadata": dict(node.metadata),
                "children": [to_json(c) for c in node.children],
            }

        return to_json(self)

    @classmethod
    def from_dict(cls, d: dict) -> Qube:
        def from_dict(d: dict) -> Iterator[Qube]:
            for k, children in d.items():
                key, values = k.split("=")
                values = values.split("/")
                if values == ["*"]:
                    values = WildcardGroup()
                else:
                    values = QEnum(values)

                yield Qube.make(
                    key=key,
                    values=values,
                    children=from_dict(children),
                )

        return Qube.root_node(list(from_dict(d)))

    def to_dict(self) -> dict:
        def to_dict(q: Qube) -> tuple[str, dict]:
            key = f"{q.key}={','.join(str(v) for v in q.values)}"
            return key, dict(to_dict(c) for c in q.children)

        return to_dict(self)[1]

    @classmethod
    def from_tree(cls, tree_str):
        lines = tree_str.splitlines()
        stack = []
        root = {}

        initial_indent = None
        for line in lines:
            if not line.strip():
                continue
            # Remove tree characters and measure indent level
            stripped = line.lstrip(" │├└─")
            indent = (len(line) - len(stripped)) // 4
            if initial_indent is None:
                initial_indent = indent
            indent = indent - initial_indent

            # Split multiple key=value parts into nested structure
            keys = [item.strip() for item in stripped.split(",")]
            current = bottom = {}
            for key in reversed(keys):
                current = {key: current}

            # Adjust the stack to current indent level
            # print(len(stack), stack)
            while len(stack) > indent:
                stack.pop()

            if stack:
                # Add to the dictionary at current stack level
                parent = stack[-1]
                key = list(current.keys())[0]
                if key in parent:
                    raise ValueError(
                        f"This function doesn't yet support reading in uncompressed trees, repeated key is {key}"
                    )
                parent[key] = current[key]
            else:
                # Top level
                key = list(current.keys())[0]
                if root:
                    raise ValueError(
                        f"This function doesn't yet support reading in uncompressed trees, repeated key is {key}"
                    )
                root = current[key]

            # Push to the stack
            stack.append(bottom)

        return cls.from_dict(root)

    @classmethod
    def empty(cls) -> Qube:
        return Qube.root_node([])

    def __str_helper__(self, depth=None, name=None) -> str:
        node = (
            dataclasses.replace(
                self,
                data=RootNodeData(key=name, values=self.values, metadata=self.metadata),
            )
            if name is not None
            else self
        )
        return "".join(node_tree_to_string(node=node, depth=depth))

    def __str__(self):
        return self.__str_helper__()

    def print(self, depth=None, name: str | None = None):
        print(self.__str_helper__(depth=depth, name=name))

    def html(self, depth=2, collapse=True, name: str | None = None) -> HTML:
        node = (
            dataclasses.replace(
                self,
                data=RootNodeData(key=name, values=self.values, metadata=self.metadata),
            )
            if name is not None
            else self
        )
        return HTML(node_tree_to_html(node=node, depth=depth, collapse=collapse))

    def _repr_html_(self) -> str:
        return node_tree_to_html(self, depth=2, collapse=True)

    # Allow "key=value/value" / qube to prepend keys
    def __rtruediv__(self, other: str) -> Qube:
        key, values = other.split("=")
        values_enum = QEnum((values.split("/")))
        return Qube.root_node([Qube.make(key, values_enum, self.children)])

    def __or__(self, other: Qube) -> Qube:
        return set_operations.operation(
            self, other, set_operations.SetOperation.UNION, type(self)
        )

    def __and__(self, other: Qube) -> Qube:
        return set_operations.operation(
            self, other, set_operations.SetOperation.INTERSECTION, type(self)
        )

    def __sub__(self, other: Qube) -> Qube:
        return set_operations.operation(
            self, other, set_operations.SetOperation.DIFFERENCE, type(self)
        )

    def __xor__(self, other: Qube) -> Qube:
        return set_operations.operation(
            self, other, set_operations.SetOperation.SYMMETRIC_DIFFERENCE, type(self)
        )

    def leaves(self) -> Iterable[dict[str, str]]:
        for value in self.values:
            if not self.children:
                yield {self.key: value}
            for child in self.children:
                for leaf in child.leaves():
                    if self.key != "root":
                        yield {self.key: value, **leaf}
                    else:
                        yield leaf

    def leaf_nodes(self) -> "Iterable[tuple[dict[str, str], Qube]]":
        for value in self.values:
            if not self.children:
                yield ({self.key: value}, self)
            for child in self.children:
                for leaf in child.leaf_nodes():
                    if self.key != "root":
                        yield ({self.key: value, **leaf[0]}, leaf[1])
                    else:
                        yield leaf

    def leaves_with_metadata(
        self, indices=()
    ) -> Iterator[tuple[dict[str, str], dict[str, str | np.ndarray]]]:
        if self.key == "root":
            for c in self.children:
                yield from c.leaves_with_metadata(indices=())
            return

        for index, value in enumerate(self.values):
            # print(self.key, index, indices, value)
            # print({k: np.shape(v) for k, v in self.metadata.items()})
            indexed_metadata = {
                k: vs[indices + (index,)] for k, vs in self.metadata.items()
            }
            indexed_metadata = {
                k: v.item() if v.shape == () else v for k, v in indexed_metadata.items()
            }
            if not self.children:
                yield {self.key: value}, indexed_metadata

            for child in self.children:
                for leaf, metadata in child.leaves_with_metadata(
                    indices=indices + (index,)
                ):
                    if self.key != "root":
                        yield {self.key: value, **leaf}, metadata | indexed_metadata
                    else:
                        yield leaf, metadata

    def datacubes(self) -> Iterable[dict[str, Any | list[Any]]]:
        def to_list_of_cubes(node: Qube) -> Iterable[dict[str, Any | list[Any]]]:
            if node.key == "root":
                for c in node.children:
                    yield from to_list_of_cubes(c)

            if not node.children:
                yield {node.key: list(node.values)}

            for c in node.children:
                for sub_cube in to_list_of_cubes(c):
                    yield {node.key: list(node.values)} | sub_cube

        return to_list_of_cubes(self)

    def __getitem__(self, args) -> Qube:
        if isinstance(args, str):
            specifiers = args.split(",")
            current = self
            for specifier in specifiers:
                key, values_str = specifier.split("=")
                values = values_str.split("/")
                for c in current.children:
                    if c.key == key and set(values) == set(c.values):
                        current = c
                        break
                else:
                    raise KeyError(
                        f"Key '{key}' not found in children of '{current.key}', available keys are {[c.key for c in current.children]}"
                    )
            return Qube.root_node(current.children)

        elif isinstance(args, tuple) and len(args) == 2:
            key, value = args
            for c in self.children:
                if c.key == key and value in c.values:
                    return Qube.root_node(c.children)
            raise KeyError(f"Key {key} not found in children of {self.key}")
        else:
            raise ValueError("Unknown key type")

    @cached_property
    def n_leaves(self) -> int:
        # This line makes the equation q.n_leaves + r.n_leaves == (q | r).n_leaves true is q and r have no overlap
        if self.key == "root" and not self.children:
            return 0
        return len(self.values) * (
            sum(c.n_leaves for c in self.children) if self.children else 1
        )

    @cached_property
    def n_nodes(self) -> int:
        if self.key == "root" and not self.children:
            return 0
        return 1 + sum(c.n_nodes for c in self.children)

    def transform(self, func: "Callable[[Qube], Qube | Iterable[Qube]]") -> Qube:
        """
        Call a function on every node of the Qube, return one or more nodes.
        If multiple nodes are returned they each get a copy of the (transformed) children of the original node.
        Any changes to the children of a node will be ignored.
        """

        def transform(node: Qube) -> list[Qube]:
            children = tuple(sorted(cc for c in node.children for cc in transform(c)))
            new_nodes = func(node)
            if isinstance(new_nodes, Qube):
                new_nodes = [new_nodes]

            return [new_node.replace(children=children) for new_node in new_nodes]

        children = tuple(cc for c in self.children for cc in transform(c))
        return self.replace(children=children)

    def remove_by_key(self, keys: str | list[str]):
        _keys: list[str] = keys if isinstance(keys, list) else [keys]

        def remove_key(node: Qube) -> Qube:
            children: list[Qube] = []
            for c in node.children:
                if c.key in _keys:
                    grandchildren = tuple(sorted(remove_key(cc) for cc in c.children))
                    grandchildren = remove_key(Qube.root_node(grandchildren)).children
                    children.extend(grandchildren)
                else:
                    children.append(remove_key(c))

            return node.replace(children=tuple(sorted(children)))

        return remove_key(self).compress()

    def convert_dtypes(self, converters: dict[str, Callable[[Any], Any]]):
        def convert(node: Qube) -> Qube:
            if node.key in converters:
                converter = converters[node.key]
                new_node = node.replace(values=QEnum(map(converter, node.values)))
                return new_node
            return node

        return self.transform(convert)

    def select(
        self,
        selection: dict[str, str | list[str] | Callable[[Any], bool]],
        mode: Literal["strict", "relaxed"] = "relaxed",
        prune=True,
        consume=False,
    ) -> Qube:
        # Find any bare str values and replace them with [str]
        _selection: dict[str, list[str] | Callable[[Any], bool]] = {}
        for k, v in selection.items():
            if isinstance(v, list):
                _selection[k] = v
            elif callable(v):
                _selection[k] = v
            else:
                _selection[k] = [v]

        def not_none(xs):
            return tuple(x for x in xs if x is not None)

        def select(
            node: Qube, selection: dict[str, list[str] | Callable[[Any], bool]]
        ) -> Qube | None:
            # If this node has no children but there are still parts of the request
            # that have not been consumed, then prune this whole branch
            if consume and not node.children and selection:
                return None

            # If the key isn't in the selection then what we do depends on the mode:
            # In strict mode we just stop here
            # In next_level mode we include the next level down so you can tell what keys to add next
            # In relaxed mode we skip the key if it't not in the request and carry on
            if node.key not in selection:
                if mode == "strict":
                    return None

                # If this node doesn't exist in the
                elif mode == "next_level":
                    return node.replace(
                        children=(),
                        metadata=self.metadata | {"is_leaf": not bool(self.children)},
                    )

                elif mode == "relaxed":
                    pass
                else:
                    raise ValueError(f"Unknown mode argument {mode}")

            # If the key IS in the selection then check if the values match
            if node.key in _selection:
                # If the key is specified, check if any of the values match
                selection_criteria = _selection[node.key]
                if callable(selection_criteria):
                    values = QEnum((c for c in node.values if selection_criteria(c)))
                elif isinstance(selection_criteria, list):
                    values = QEnum((c for c in selection_criteria if c in node.values))
                else:
                    raise ValueError(f"Unknown selection type {selection_criteria}")

                # Here modes don't matter because we've explicitly filtered on this key and found nothing
                if not values:
                    return None

                node = node.replace(values=values)

            if consume:
                selection = {k: v for k, v in selection.items() if k != node.key}

            # Prune nodes that had had all their children pruned
            new_children = not_none(select(c, selection) for c in node.children)
            # if node.key == "dataset": print(prune, [(c.key, c.values.values) for c in node.children], [c.key for c in new_children])

            if prune and node.children and not new_children:
                return None

            return node.replace(
                children=new_children,
                metadata=dict(self.metadata) | {"is_leaf": not bool(new_children)},
            )

        return self.replace(
            children=not_none(select(c, _selection) for c in self.children)
        )

    def span(self, key: str) -> list[str]:
        """
        Search the whole tree for any value that a given key takes anywhere.
        """
        this = set(self.values) if self.key == key else set()
        return sorted(this | set(v for c in self.children for v in c.span(key)))

    def axes(self) -> dict[str, set[str]]:
        """
        Return a dictionary of all the spans of the keys in the qube.
        """
        axes = defaultdict(set)
        for c in self.children:
            for k, v in c.axes().items():
                axes[k].update(v)
        if self.key != "root":
            axes[self.key].update(self.values)
        return dict(axes)

    @cached_property
    def structural_hash(self) -> int:
        """
        This hash takes into account the key, values and children's key values recursively.
        Because nodes are immutable, we only need to compute this once.
        """

        def hash_node(node: Qube) -> int:
            return hash(
                (node.key, node.values, tuple(c.structural_hash for c in node.children))
            )

        return hash_node(self)

    def compress(self) -> Qube:
        """
        This method is quite computationally heavy because of trees like this:
        root, class=d1, generation=1
        ├── time=0600, many identical keys, param=8,78,79
        ├── time=0600, many identical keys, param=8,78,79
        └── time=0600, many identical keys, param=8,78,79
        This tree compresses dow n

        """

        def union(a: Qube, b: Qube) -> Qube:
            b = type(self).root_node(children=(b,))
            out = set_operations.operation(
                a, b, set_operations.SetOperation.UNION, type(self)
            )
            return out

        new_children = [c.compress() for c in self.children]
        if len(new_children) > 1:
            new_children = list(
                functools.reduce(union, new_children, Qube.empty()).children
            )

        return self.replace(children=tuple(sorted(new_children)))



from dataclasses import dataclass, field

character = str


@dataclass(unsafe_hash=True)
class TrieNode:
    parent: "TrieNode | None"
    parent_char: character
    children: dict[character, "TrieNode"] = field(default_factory=dict)


@dataclass
class Trie:
    root: TrieNode = field(default_factory=lambda: TrieNode(None, ""))
    reverse_lookup: dict[int, TrieNode] = field(default_factory=dict)

    def insert(self, word: str):
        node = self.root
        for char in word:
            if char not in node.children:
                new_node = TrieNode(node, char)
                node.children[char] = new_node

            node = node.children[char]

        n_id = id(node)
        if n_id not in self.reverse_lookup:
            self.reverse_lookup[n_id] = node

        return n_id

    def lookup_by_id(self, n_id: int):
        leaf_node = self.reverse_lookup[n_id]
        string = []
        while leaf_node.parent is not None:
            string.append(leaf_node.parent_char)
            leaf_node = leaf_node.parent

        return "".join(reversed(string))



from qubed import Qube

d = {
    "class=od": {
        "expver=0001": {"param=1": {}, "param=2": {}},
        "expver=0002": {"param=1": {}, "param=2": {}},
    },
    "class=rd": {
        "expver=0001": {"param=1": {}, "param=2": {}, "param=3": {}},
        "expver=0002": {"param=1": {}, "param=2": {}},
    },
}
q = Qube.from_dict(d).compress()

as_string = """
root
├── class=od, expver=0001/0002, param=1/2
└── class=rd
    ├── expver=0001, param=1/2/3
    └── expver=0002, param=1/2
""".strip()

as_html = """
<details open data-path="root"><summary class="qubed-node">root</summary><span class="qubed-node leaf" data-path="class=od,expver=0001/0002,param=1/2">├── class=od, expver=0001/0002, param=1/2</span><details open data-path="class=rd"><summary class="qubed-node">└── class=rd</summary><span class="qubed-node leaf" data-path="expver=0001,param=1/2/3">    ├── expver=0001, param=1/2/3</span><span class="qubed-node leaf" data-path="expver=0002,param=1/2">    └── expver=0002, param=1/2</span></details></details>
""".strip()


def test_string():
    assert str(q).strip() == as_string


def test_html():
    assert as_html in q._repr_html_()



from dataclasses import dataclass, field
from typing import Mapping

import numpy as np
from frozendict import frozendict

from .value_types import ValueGroup


@dataclass(frozen=False, eq=True, order=True, unsafe_hash=True)
class NodeData:
    key: str
    values: ValueGroup
    metadata: Mapping[str, np.ndarray] = field(
        default_factory=frozendict, compare=False
    )

    def summary(self) -> str:
        return f"{self.key}={self.values.summary()}" if self.key != "root" else "root"


@dataclass(frozen=False, eq=True, order=True)
class RootNodeData(NodeData):
    "Helper class to print a custom root name"

    def summary(self) -> str:
        return self.key



import json
import time

import click
import psutil
from rich.console import Console
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.spinner import Spinner
from rich.text import Text

from qubed import Qube
from qubed.convert import parse_fdb_list

console = Console(stderr=True)
process = psutil.Process()

PRINT_INTERVAL = 0.25


@click.group()
def main():
    """Command-line tool for working with trees."""
    pass


@main.command()
@click.option(
    "--input",
    type=click.File("r"),
    default="-",
    help="Specify the input file (default: standard input).",
)
@click.option(
    "--output",
    type=click.File("w"),
    default="-",
    help="Specify the output file (default: standard output).",
)
@click.option(
    "--from",
    "from_format",
    type=click.Choice(["fdb", "mars"]),
    default="fdb",
    help="Specify the input format: fdb (fdb list --porcelain) or mars (mars list).",
)
@click.option(
    "--to",
    "to_format",
    type=click.Choice(["text", "html", "json"]),
    default="text",
    help="Specify the output format: text, html, json.",
)
def convert(input, output, from_format, to_format):
    """Convert trees from one format to another."""
    q = Qube.empty()
    t = time.time()
    i0 = 0
    n0 = 0
    depth = 5
    log = Text()
    summary = Layout()
    summary.split_column(
        Layout(name="upper"),
        Layout(name="qube"),
    )
    summary["upper"].split_row(
        Layout(name="performance"),
        Layout(log, name="log"),
    )
    spinner = Spinner("aesthetic", text="Performance", speed=0.3)

    with Live(summary, auto_refresh=False, transient=True, console=console) as live:
        for i, datacube in enumerate(parse_fdb_list(input)):
            new_branch = Qube.from_datacube(datacube)
            q = q | new_branch

            if time.time() - t > PRINT_INTERVAL:
                tree = q.__str__(depth=depth)
                if tree.count("\n") > 20:
                    depth -= 1
                if tree.count("\n") < 5:
                    depth += 1

                summary["performance"].update(
                    Panel(
                        Text.assemble(
                            f"The Qube has {q.n_leaves} leaves and {q.n_nodes} internal nodes so far.\n",
                            f"{(i - i0) / (time.time() - t) / PRINT_INTERVAL:.0f} lines per second.  ",
                            f"{(q.n_leaves - n0) / (time.time() - t):.0f} leaves per second.\n",
                            f"Memory usage: {process.memory_info().rss / 1024 / 1024:.0f} MB\n",
                        ),
                        title=spinner.render(time.time()),
                        border_style="blue",
                    )
                )
                summary["qube"].update(
                    Panel(tree, title=f"Qube (depth {depth})", border_style="blue")
                )
                summary["log"].update(
                    Panel(
                        f"{datacube}", border_style="blue", title="Last Datacube Added"
                    )
                )
                live.refresh()
                i0 = i
                n0 = q.n_leaves
                t = time.time()

    if to_format == "text":
        output_content = str(q)
    elif to_format == "json":
        output_content = json.dumps(q.to_json())
    elif to_format == "html":
        output_content = q.html().html
    else:
        output_content = str(q)

    output.write(output_content)


if __name__ == "__main__":
    main()



import re
import subprocess
from pathlib import Path

CARGO_TOML_PATH = Path("Cargo.toml")


# Get the latest Git tag and strip the leading 'v' if present
def get_git_version():
    try:
        version = subprocess.check_output(
            ["git", "describe", "--tags", "--always"], text=True
        ).strip()
        version = re.sub(r"^v", "", version)  # Remove leading 'v'
        return version
    except subprocess.CalledProcessError:
        raise RuntimeError(
            "Failed to get Git tag. Make sure you have at least one tag in the repository."
        )


# Update version in Cargo.toml
def update_cargo_version(new_version):
    cargo_toml = CARGO_TOML_PATH.read_text()

    # Replace version in [package] section
    updated_toml = re.sub(
        r'^version = "[^"]+"',
        f'version = "{new_version}"',
        cargo_toml,
        flags=re.MULTILINE,
    )

    CARGO_TOML_PATH.write_text(updated_toml)


if __name__ == "__main__":
    version = get_git_version()
    print(f"Parsed version: {version}")

    update_cargo_version(version)
    print(f"Updated Cargo.toml with version: {version}")



import json
from pathlib import Path

from tree_traverser import CompressedTree

data_path = Path("./config/climate-dt/compressed_tree.json")
# Print size of file
print(f"climate dt compressed tree: {data_path.stat().st_size // 1e6:.1f} MB")

print("Opening json file")
compressed_tree = CompressedTree.load(data_path)

print(compressed_tree.to_json())

print("Outputting compressed tree ecmwf style")
with open("config/climate-dt/new_format.json", "w") as f:
    json.dump(compressed_tree.to_json(), f)



import json
from collections import defaultdict

from qubed import Qube

metadata = json.load(open("raw_anemoi_metadata.json"))

predicted_indices = [
    *metadata["data_indices"]["data"]["output"]["prognostic"],
    *metadata["data_indices"]["data"]["output"]["diagnostic"],
]
variables = metadata["dataset"]["variables"]
variables = [variables[i] for i in predicted_indices]

# print('Raw Model Variables:', variables)

# Split variables between pressure and surface
surface_variables = [v for v in variables if "_" not in v]

# Collect the levels for each pressure variable
level_variables = defaultdict(list)
for v in variables:
    if "_" in v:
        variable, level = v.split("_")
        level_variables[variable].append(int(level))

# print(level_variables)

model_tree = Qube.empty()

for variable, levels in level_variables.items():
    model_tree = model_tree | Qube.from_datacube(
        {
            "levtype": "pl",
            "param": variable,
            "level": levels,
        }
    )

for variable in surface_variables:
    model_tree = model_tree | Qube.from_datacube(
        {
            "levtype": "sfc",
            "param": variable,
        }
    )

print(model_tree.to_json())



from __future__ import annotations

import random
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from .Qube import Qube


@dataclass(frozen=True)
class HTML:
    html: str

    def _repr_html_(self):
        return self.html


def summarize_node(
    node: Qube, collapse=False, max_summary_length=50, **kwargs
) -> tuple[str, str, Qube]:
    """
    Extracts a summarized representation of the node while collapsing single-child paths.
    Returns the summary string and the last node in the chain that has multiple children.
    """
    summaries = []
    paths = []

    while True:
        summary = node.summary(**kwargs)
        if "is_leaf" in node.metadata and node.metadata["is_leaf"]:
            summary += " 🌿"
        paths.append(summary)
        if len(summary) > max_summary_length:
            summary = summary[:max_summary_length] + "..."
        summaries.append(summary)
        if not collapse:
            break

        # Move down if there's exactly one child, otherwise stop
        if len(node.children) != 1:
            break
        node = node.children[0]

    return ", ".join(summaries), ",".join(paths), node


def node_tree_to_string(node: Qube, prefix: str = "", depth=None) -> Iterable[str]:
    summary, path, node = summarize_node(node)

    if depth is not None and depth <= 0:
        yield summary + " - ...\n"
        return
    # Special case for nodes with only a single child, this makes the printed representation more compact
    elif len(node.children) == 1:
        yield summary + ", "
        yield from node_tree_to_string(node.children[0], prefix, depth=depth)
        return
    else:
        yield summary + "\n"

    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        yield prefix + connector
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from node_tree_to_string(
            child, prefix + extension, depth=depth - 1 if depth is not None else None
        )


def _node_tree_to_html(
    node: Qube, prefix: str = "", depth=1, connector="", **kwargs
) -> Iterable[str]:
    summary, path, node = summarize_node(node, **kwargs)

    if len(node.children) == 0:
        yield f'<span class="qubed-node leaf" data-path="{path}">{connector}{summary}</span>'
        return
    else:
        open = "open" if depth > 0 else ""
        yield f'<details {open} data-path="{path}"><summary class="qubed-node">{connector}{summary}</summary>'

    for index, child in enumerate(node.children):
        connector = "└── " if index == len(node.children) - 1 else "├── "
        extension = "    " if index == len(node.children) - 1 else "│   "
        yield from _node_tree_to_html(
            child,
            prefix + extension,
            depth=depth - 1,
            connector=prefix + connector,
            **kwargs,
        )
    yield "</details>"


def node_tree_to_html(
    node: Qube, depth=1, include_css=True, include_js=True, css_id=None, **kwargs
) -> str:
    if css_id is None:
        css_id = f"qubed-tree-{random.randint(0, 1000000)}"

    # It's ugle to use an f string here because css uses {} so much so instead
    # we use CSS_ID as a placeholder and replace it later
    css = """
        <style>
        pre#CSS_ID {
            font-family: monospace;
            white-space: pre;
            font-family: SFMono-Regular,Menlo,Monaco,Consolas,Liberation Mono,Courier New,Courier,monospace;
            font-size: 12px;
            line-height: 1.4;

            details {
                margin-left: 0;
            }

            .qubed-node a {
                margin-left: 10px;
                text-decoration: none;
            }

            summary {
                list-style: none;
                cursor: pointer;
                text-overflow: ellipsis;
                overflow: hidden;
                text-wrap: nowrap;
                display: block;
            }

            summary:hover,span.leaf:hover {
                background-color: #f0f0f0;
            }

            details > summary::after {
                content: ' ▲';
            }

            details:not([open]) > summary::after {
                content: " ▼";
            }

            .leaf {
                text-overflow: ellipsis;
                overflow: hidden;
                text-wrap: nowrap;
                display: block;
            }

            summary::-webkit-details-marker {
              display: none;
              content: "";
            }

        }
        </style>
        """.replace("CSS_ID", css_id)

    # This js snippet copies the path of a node to the clipboard when clicked
    js = """
        <script type="module" defer>
        async function nodeOnClick(event) {
            if (!event.altKey) return;
            event.preventDefault();
            let current_element = this.parentElement;
            let paths = [];
            while (true) {
                if (current_element.dataset.path) {
                    paths.push(current_element.dataset.path);
                }
                current_element = current_element.parentElement;
                if (current_element.tagName == "PRE") break;
            }
            const path = paths.reverse().slice(1).join(",");
            await navigator.clipboard.writeText(path);
        }

        const nodes = document.querySelectorAll("#CSS_ID.qubed-node");
        nodes.forEach(n => n.addEventListener("click", nodeOnClick));
        </script>
        """.replace("CSS_ID", css_id)
    nodes = "".join(_node_tree_to_html(node=node, depth=depth, **kwargs))
    return f"{js if include_js else ''}{css if include_css else ''}<pre class='qubed-tree' id='{css_id}'>{nodes}</pre>"



from __future__ import annotations

from collections import defaultdict
from dataclasses import replace
from enum import Enum

# Prevent circular imports while allowing the type checker to know what Qube is
from typing import TYPE_CHECKING, Iterable

from frozendict import frozendict

from .node_types import NodeData
from .value_types import QEnum, ValueGroup, WildcardGroup

if TYPE_CHECKING:
    from .Qube import Qube


class SetOperation(Enum):
    UNION = (1, 1, 1)
    INTERSECTION = (0, 1, 0)
    DIFFERENCE = (1, 0, 0)
    SYMMETRIC_DIFFERENCE = (1, 0, 1)


def node_intersection(
    A: ValueGroup, B: ValueGroup
) -> tuple[ValueGroup, ValueGroup, ValueGroup]:
    if isinstance(A, QEnum) and isinstance(B, QEnum):
        set_A, set_B = set(A), set(B)
        intersection = set_A & set_B
        just_A = set_A - intersection
        just_B = set_B - intersection
        return QEnum(just_A), QEnum(intersection), QEnum(just_B)

    if isinstance(A, WildcardGroup) and isinstance(B, WildcardGroup):
        return A, WildcardGroup(), B

    # If A is a wildcard matcher then the intersection is everything
    # just_A is still *
    # just_B is empty
    if isinstance(A, WildcardGroup):
        return A, B, QEnum([])

    # The reverse if B is a wildcard
    if isinstance(B, WildcardGroup):
        return QEnum([]), A, B

    raise NotImplementedError(
        f"Fused set operations on values types {type(A)} and {type(B)} not yet implemented"
    )


def operation(A: Qube, B: Qube, operation_type: SetOperation, node_type) -> Qube:
    assert A.key == B.key, (
        "The two Qube root nodes must have the same key to perform set operations,"
        f"would usually be two root nodes. They have {A.key} and {B.key} respectively"
    )

    assert A.values == B.values, (
        f"The two Qube root nodes must have the same values to perform set operations {A.values = }, {B.values = }"
    )

    # Group the children of the two nodes by key
    nodes_by_key: defaultdict[str, tuple[list[Qube], list[Qube]]] = defaultdict(
        lambda: ([], [])
    )
    for node in A.children:
        nodes_by_key[node.key][0].append(node)
    for node in B.children:
        nodes_by_key[node.key][1].append(node)

    new_children: list[Qube] = []

    # For every node group, perform the set operation
    for key, (A_nodes, B_nodes) in nodes_by_key.items():
        new_children.extend(
            _operation(key, A_nodes, B_nodes, operation_type, node_type)
        )

    # Whenever we modify children we should recompress them
    # But since `operation` is already recursive, we only need to compress this level not all levels
    # Hence we use the non-recursive _compress method
    new_children = list(compress_children(new_children))

    # The values and key are the same so we just replace the children
    return A.replace(children=tuple(sorted(new_children)))


# The root node is special so we need a helper method that we can recurse on
def _operation(
    key: str, A: list[Qube], B: list[Qube], operation_type: SetOperation, node_type
) -> Iterable[Qube]:
    keep_just_A, keep_intersection, keep_just_B = operation_type.value

    # Iterate over all pairs (node_A, node_B)
    values = {}
    for node in A + B:
        values[node] = node.values

    for node_a in A:
        for node_b in B:
            # Compute A - B, A & B, B - A
            # Update the values for the two source nodes to remove the intersection
            just_a, intersection, just_b = node_intersection(
                values[node_a],
                values[node_b],
            )

            # Remove the intersection from the source nodes
            values[node_a] = just_a
            values[node_b] = just_b

            if keep_intersection:
                if intersection:
                    new_node_a = replace(
                        node_a, data=replace(node_a.data, values=intersection)
                    )
                    new_node_b = replace(
                        node_b, data=replace(node_b.data, values=intersection)
                    )
                    yield operation(new_node_a, new_node_b, operation_type, node_type)

    # Now we've removed all the intersections we can yield the just_A and just_B parts if needed
    if keep_just_A:
        for node in A:
            if values[node]:
                yield node_type.make(key, values[node], node.children)
    if keep_just_B:
        for node in B:
            if values[node]:
                yield node_type.make(key, values[node], node.children)


def compress_children(children: Iterable[Qube]) -> tuple[Qube, ...]:
    """
    Helper method tht only compresses a set of nodes, and doesn't do it recursively.
    Used in Qubed.compress but also to maintain compression in the set operations above.
    """
    # Now take the set of new children and see if any have identical key, metadata and children
    # the values may different and will be collapsed into a single node
    identical_children = defaultdict(set)
    for child in children:
        # only care about the key and children of each node, ignore values
        h = hash((child.key, tuple((cc.structural_hash for cc in child.children))))
        identical_children[h].add(child)

    # Now go through and create new compressed nodes for any groups that need collapsing
    new_children = []
    for child_set in identical_children.values():
        if len(child_set) > 1:
            child_list = list(child_set)
            node_type = type(child_list[0])
            key = child_list[0].key

            # Compress the children into a single node
            assert all(isinstance(child.data.values, QEnum) for child in child_set), (
                "All children must have QEnum values"
            )

            node_data = NodeData(
                key=str(key),
                metadata=frozendict(),  # Todo: Implement metadata compression
                values=QEnum((v for child in child_set for v in child.data.values)),
            )
            new_child = node_type(data=node_data, children=child_list[0].children)
        else:
            # If the group is size one just keep it
            new_child = child_set.pop()

        new_children.append(new_child)
    return tuple(sorted(new_children, key=lambda n: ((n.key, n.values.min()))))


def union(a: Qube, b: Qube) -> Qube:
    return operation(
        a,
        b,
        SetOperation.UNION,
        type(a),
    )



from qubed import Qube


def test_json_round_trip():
    u = Qube.from_dict(
        {
            "class=d1": {
                "dataset=climate-dt/weather-dt": {
                    "generation=1/2/3/4": {},
                },
                "dataset=another-value": {
                    "generation=1/2/3": {},
                },
            }
        }
    )
    json = u.to_json()
    assert Qube.from_json(json) == u



from pathlib import Path

from tree_traverser import CompressedTree, RefcountedDict


class CompressedTreeFixed(CompressedTree):
    @classmethod
    def from_json(cls, data: dict):
        c = cls({})
        c.cache = {}
        ca = data["cache"]
        for k, v in ca.items():
            g = {
                k2: ca[str(v2)]["dict"][k2] if k2 in ca[str(v2)]["dict"] else v2
                for k2, v2 in v["dict"].items()
            }
            c.cache[int(k)] = RefcountedDict(g)
            c.cache[int(k)].refcount = v["refcount"]

        c.root_hash = data["root_hash"]
        c.tree = c.cache[c.root_hash]
        return c

    def reconstruct(self, max_depth=None) -> dict[str, dict]:
        "Reconstruct the tree as a normal nested dictionary"

        def reconstruct_node(h: int, depth: int) -> dict[str, dict]:
            if max_depth is not None and depth > max_depth:
                return {}
            return {
                k: reconstruct_node(v, depth=depth + 1)
                for k, v in self.cache[h].items()
            }

        return reconstruct_node(self.root_hash, depth=0)


data_path = Path("data/compressed_tree_climate_dt.json")
# Print size of file
print(f"climate dt compressed tree: {data_path.stat().st_size // 1e6:.1f} MB")

print("Opening json file")
compressed_tree = CompressedTreeFixed.load(data_path)

output_data_path = Path("data/compressed_tree_climate_dt_ecmwf_style.json")
# Print size of file

compressed_tree.save(output_data_path)

print(
    f"climate dt compressed tree ecmwf style: {output_data_path.stat().st_size // 1e6:.1f} MB"
)



from qubed import Qube


def test_leaf_conservation():
    q = Qube.from_dict(
        {
            "class=d1": {
                "dataset=climate-dt": {
                    "time=0000": {
                        "param=130/134/137/146/147/151/165/166/167/168/169": {}
                    },
                    "time=0001": {"param=130": {}},
                }
            }
        }
    )

    r = Qube.from_datacube(
        {"class": "d1", "dataset": "climate-dt", "time": "0001", "param": "134"}
    )

    assert q.n_leaves + r.n_leaves == (q | r).n_leaves



from pathlib import Path

import orjson as json
from tree_traverser.DataCubeTree import CompressedTree

data_path = Path("./config/climate-dt/new_format.json")
with data_path.open("r") as f:
    compressed_tree = CompressedTree.from_json(json.loads(f.read()))

compressed_tree = compressed_tree.guess_datatypes()

compressed_tree.print(depth=10)



from qubed import Qube


def test_simple():
    q = Qube.from_tree("""
    root, frequency=6:00:00
    ├── levtype=pl, param=t, levelist=850, threshold=-2/-4/-8/2/4/8
    └── levtype=sfc
        ├── param=10u/10v, threshold=10/15
        ├── param=2t, threshold=273.15
        └── param=tp, threshold=0.1/1/10/100/20/25/5/50
    """)

    r = Qube.from_dict(
        {
            "frequency=6:00:00": {
                "levtype=pl": {
                    "param=t": {"levelist=850": {"threshold=-8/-4/-2/2/4/8": {}}}
                },
                "levtype=sfc": {
                    "param=10u/10v": {"threshold=10/15": {}},
                    "param=2t": {"threshold=273.15": {}},
                    "param=tp": {"threshold=0.1/1/5/10/20/25/50/100": {}},
                },
            },
        }
    )

    assert q == r


def test_simple_2():
    models = Qube.from_datacube(
        dict(
            param="10u/10v/2d/2t/cp/msl/skt/sp/tcw/tp".split("/"),
            threshold="*",
            levtype="sfc",
            frequency="6:00:00",
        )
    ) | Qube.from_datacube(
        dict(
            param="q/t/u/v/w/z".split("/"),
            threshold="*",
            levtype="pl",
            level="50/100/150/200/250/300/400/500/600/700/850".split("/"),
            frequency="6:00:00",
        )
    )

    models2 = Qube.from_tree("""
    models
    ├── param=10u/10v/2d/2t/cp/msl/skt/sp/tcw/tp, threshold=*, levtype=sfc, frequency=6:00:00
    └── param=q/t/u/v/w/z, threshold=*, levtype=pl, level=100/150/200/250/300/400/50/500/600/700/850, frequency=6:00:00
    """)

    assert models == models2
